import os
import email

import docutils.core
import webob

from webob import exc
from webob.dec import wsgify


#from wsgilite.framework.method import HttpResource
from wsgilite.framework.cn import vary


class RstResource(object):
    r"""

    convert a resource in reStructuredText (a-la docutils) into any of a
    variety of other, useful formats.

    To use this class, you must provide an variant of the `render()` method for
    the `text/x-rst` media type, (the rStructuredText mime type).  which
    roughly looks like

    >>> import textwrap
    >>> class RstTestResource(RstResource):
    ...     document = textwrap.dedent('''
    ...     Hello World
    ...     ===========
    ...
    ...     - foo
    ...     - bar
    ...     - baz
    ...     ''').strip()
    ...
    ...     @RstResource.render(accept='text/x-rst')
    ...     @wsgify
    ...     def render(self, req):
    ...         return webob.Response(self.document)

    """

    @wsgify
    def __call__(self, req):
        print "RstResource --> render"
        return req.get_response(self.render(req))

    @vary(default_media='text/html')
    @wsgify
    def render(self, req):
        # obtain the reStructuredText version:
        print "RstResource.render --> text/x-rst"
        rstresponse = req.get_response(self.render.media['text/x-rst'])
        rst = rstresponse.text
        return webob.Response(docutils.core.publish_string(rst, writer_name='html'))


#FIXME: this repeats some of the stuff in static.StaticFile...
class RstFile(RstResource):
    def __init__(self, path):
        self.path = path

    @RstResource.render(accept='text/x-rst')
    @wsgify
    def render(self, req):
        if not os.path.isfile(self.path):
            raise exc.HTTPNotFound

        resp = webob.Response()
        resp.content_type = 'text/x-rst'

        file_info = os.stat(self.path)
        resp.date = email.utils.formatdate(file_info.st_mtime)
        resp.body_file = open(self.path)
        return resp



#! /usr/bin/env python
try:
    import __builtin__ as builtins
    assert builtins  # silence pyflakes
except ImportError:
    import builtins

try:
    basestring
except NameError:
    basestring = unicode = str

import inspect
import logging
import os
import sys
import threading
import time

#TODO: Add the funny game/rpg elements back
#TODO: Add syntax checking
#TODO: use inotify instead of stat (where available)
#TODO: trim python std libs (optionally?)

__all__ = ['assassinate']
LOGGER = logging.getLogger(__name__)


def guesswatchfile(thing):
    if inspect.ismodule(thing):
        try:
            thingsourcefile = inspect.getsourcefile(thing)
            if not os.path.exists(thingsourcefile):
                raise TypeError("not really a file", thingsourcefile)
            return thingsourcefile

        except TypeError:
            LOGGER.debug("cannot watch %r, no source file", thing)
            return None
    if thing in sys.modules:
        return guesswatchfile(sys.modules[thing])
    elif isinstance(thing, basestring) and os.path.exists(thing):
        return thing
    raise ValueError("Couldn't figure out what to watch", thing)


def check_syntax(path):
    try:
        compile(open(path).read(), path, "exec")
        return True
    except SyntaxError:
        LOGGER.exception("target did not compile %r", path)
        return os.path.splitext(path, '.py')


def restart_execl(argv=list(sys.argv),
                  closerange=(3, 1000)):
    LOGGER.debug("restart_execl(%r, %r)", argv, closerange)
    os.closerange(*closerange)
    os.execl(argv[0], *argv),


def assassinate(watch_files,
                sleep_time=1,
                restart=restart_execl,
                _reentry=[],
                _real_watchfiles={},
                _lock=threading.Lock()):

    # watch some files
    with _lock:
        for watch_thing in watch_files:
            LOGGER.debug("assasinate(%r)", watch_thing)
            watch = guesswatchfile(watch_thing)
            if watch is None or watch in _real_watchfiles:
                continue
            LOGGER.debug("watching %s", watch)
            _real_watchfiles[watch] = os.stat(watch).st_mtime

        # if this has been called before, bail out
        if _reentry:
            return
        _reentry.append(True)

    # monkey patch imports
    real_import = builtins.__import__

    def assassin_import(name, globals={}, locals={}, fromlist=[], level=-1):
        peek_keys = sys.modules.keys()
        importedModule = real_import(name, globals, locals, fromlist, level)
        if importedModule.__name__ not in peek_keys:
            assassinate([importedModule])
        return importedModule

    LOGGER.debug("replacing __builtin__.__import__")
    builtins.__import__ = assassin_import

    def assassin():
        while True:
            with _lock:
                for watch, mtime in _real_watchfiles.items():
                    # retry a couple times in case the file is being moved
                    # about non-atomically
                    for tries in range(2, -1, -1):
                        try:
                            new_mtime = os.stat(watch).st_mtime
                            if new_mtime != mtime:
                                LOGGER.info("Restarting on file change, %r", watch)
                                if check_syntax(watch):
                                    restart()
                                else:
                                    _real_watchfiles[watch] = new_mtime
                        except OSError:
                            if tries:
                                time.sleep(0.3)
                            else:
                                raise

            time.sleep(sleep_time)

    athread = threading.Thread(target=assassin)
    athread.daemon = True
    athread.start()



from webob.dec import wsgify
from webob import Response


@wsgify
def echo_app(req):
    resp = Response(body=str(req))
    resp.content_type = 'text/plain'
    return resp


@wsgify
def echo_status(req):
    status = req.path_info.lstrip('/')
    resp = Response()
    resp.status = status
    return resp



import email
import mimetypes
import os
import urllib

import webob
from webob.dec import wsgify

from wsgilite.framework.method import HttpResource
from wsgilite.framework.method import httpmethod


__all__ = [
    'IndexFile',
    'StaticDir',
    'StaticFile',
]


class StaticFile(HttpResource):

    not_found = webob.Response(status=404)

    def __init__(self, path, content_type=None, content_encoding=None):
        self.path = path
        if content_type is None:
            content_type, content_encoding = mimetypes.guess_type(self.path)
        self.content_type = content_type
        self.content_encoding = content_encoding

    @httpmethod
    def HEAD(self, req):
        if not os.path.isfile(self.path):
            return self.not_found

        resp = webob.Response()
        resp.content_type = self.content_type
        resp.content_encoding = self.content_encoding

        file_info = os.stat(self.path)
        resp.date = email.utils.formatdate(file_info.st_mtime)
        return resp

    @httpmethod
    def GET(self, req):
        resp = req.get_response(self.HEAD)
        if 200 <= resp.status_int < 300:
            resp.body_file = open(self.path)
        return resp


class IndexFile(StaticFile):
    index_filename = 'index.html'

    @property
    def path(self):
        return os.path.join(self.real_path, self.index_filename)

    @path.setter
    def path(self, value):
        self.real_path = value


class StaticDir(object):
    """
    static file directory serving app.
    """
    static_app = StaticFile
    index_app = IndexFile
    not_found = webob.Response(status=404)

    def __init__(self, path='.'):
        self.path = path

    @wsgify
    def __call__(self, req):
        path = urllib.unquote_plus(req.path_info) or '/'
        # assert path.startswith('/')
        baseurl = os.path.normpath(path)
        abspath = os.path.join(self.path, baseurl[1:]).rstrip('/')

        if os.path.isfile(abspath):
            return req.get_response(self.static_app(abspath))  # (environ, start_response)

        elif os.path.isdir(abspath):
            return req.get_response(self.index_app(abspath, baseurl))  # (environ, start_response)

        else:
            return req.get_response(self.not_found)  # (environ, start_response)



import logging
from webob.dec import wsgify
from webob import Request
LOGGER = logging.getLogger(__name__)


def sqlalchemy_middlware(sessionfactory):
    def _wrapped(app):
        return SqlalchemyMiddleware(app, sessionfactory)
    return _wrapped


class SqlalchemyMiddleware(object):
    r"""
    A sqlalchemy session middlware.

    this package has no dependency on sqlalchemy or any database, it requires
    injection of a session factory (for instance, sqlalchemy.orm.sessionmaker).
    which has the side benefit of making it a little clearer what actually
    happens during each request, by substituting a mock session:

        >>> import minimock
        >>> from webob import Response
        >>> Session = minimock.Mock("Session")
        >>> Session.mock_returns = minimock.Mock("session")
        >>> @sqlalchemy_middlware(Session)
        ... @wsgify
        ... def app(req):
        ...     if req.GET.get('session'):
        ...         session = SqlalchemyMiddleware.get_session(req)
        ...     if req.GET.get('except'):
        ...         1/0
        ...     return Response(status=req.GET.get('status'))
        ...

    If the session is never requested, it doesn't get created either.

        >>> print Request.blank("").get_response(app)
        200 OK
        Content-Type: text/html; charset=UTF-8
        Content-Length: 0

    The session is closed at the end of the request (before consuming the
    returned app_iter)

        >>> print Request.blank("/?session=yes").get_response(app)
        Called Session()
        Called session.close()
        200 OK
        Content-Type: text/html; charset=UTF-8
        Content-Length: 0

    Errors propagate, but are trapped as they pass through the middlware for
    the purposes of cleaning up the session:

        >>> try:
        ...     print Request.blank("/?except=yes&session=yes") \
        ...                  .get_response(app)
        ... except ZeroDivisionError:
        ...     print "it died"
        ...
        Called Session()
        Called session.rollback()
        Called session.close()
        it died

    As before, if the session isn't used, then there's no session at all;  that
    goes for cleanup as well.

        >>> print Request.blank("/?except=yes").get_response(app)
        Traceback (most recent call last):
            1/0
        ZeroDivisionError: integer division or modulo by zero


    """

    # TODO: multi-database semantics
    #   change get_session(req) to get_session(req, "db-id") (optional)
    # TODO: autocommit on response (matches common expectations)

    def __init__(self, application, session_factory):
        self.application = application
        self.session_factory = session_factory

    @classmethod
    def get_session(cls, req=None, environ=None):
        """
        Create a new session and attach it to the supplied Request, or return
        the already attached session if it exists
        """
        if req is None:
            req = Request(environ)
        try:
            return req.sqlalchemy_middlware_session
        except AttributeError:
            pass
        self = req.sqlalchemy_middleware_instance
        req.sqlalchemy_middlware_session = session = self.session_factory()
        LOGGER.debug("sa.mw creating new session")
        return session

    @classmethod
    def rollback(cls, req=None, environ=None):
        """
        Roll back the current transaction on the supplied Request, or do
        nothing if there is no current session.
        """
        if req is None:
            req = Request(environ)
        try:
            session = req.sqlalchemy_middlware_session
        except AttributeError:
            LOGGER.debug("sa.mw no session to rollback")
            return
        LOGGER.debug("sa.mw session.rollback")
        session.rollback()

    @classmethod
    def close(cls, req=None, environ=None):
        """
        Close the current transaction on the supplied Request, or do
        nothing if there is no current session.
        """
        if req is None:
            req = Request(environ)
        try:
            session = req.sqlalchemy_middlware_session
        except AttributeError:
            LOGGER.debug("sa.mw no session to close")
            return
        LOGGER.debug("sa.mw session.close")
        session.close()
        del req.sqlalchemy_middlware_session

    @wsgify
    def __call__(self, req):
        req.sqlalchemy_middleware_instance = self
        app = self.application
        try:
            resp = req.get_response(app)
            return resp
        except:
            LOGGER.debug("sa.mw rollback forced on exception in %r()",
                         app, exc_info=True)
            self.rollback(req)
            raise
        finally:
            self.close(req)
            del req.sqlalchemy_middleware_instance



from webob.dec import wsgify

from wsgilite.apps import static
from wsgilite.extras import restructured


class DynamicFile(object):
    raw = static.StaticFile
    rst = restructured.RstFile

    def __init__(self, path):
        self.path = path

    @wsgify
    def __call__(self, req):
        if req.GET.get('raw'):
            raw = self.raw(self.path)
            return req.get_response(raw)
        elif req.GET.get('rst'):
            print "DynamicFileApp --> rst", self.path
            rst = self.rst(self.path)
            return req.get_response(rst)
        else:
            raw = self.raw(self.path)
            return req.get_response(raw)


class DynamicDir(static.StaticDir):
    static_app = DynamicFile



from webob.dec import wsgify
from webob import Response


def set_httpmethod(wrapped):
    """
    inform HttpResource that the decorated function is a wsgi callable that
    represents a specific HTTP method.
    """
    wrapped._httpmethod = True
    return wrapped


class httpmethod(wsgify):
    """
    inform HttpResource that the decorated function takes a webob.Request and
    returns a wsgi callable (eg, returns ResponseEx instance) for a specific
    HTTP method.
    """
    _httpmethod = True


# TODO: implement OPTIONS;
class HttpResource(object):
    """
        >>> from webob import Request, Response
        >>> class HTCPTP(HttpResource):
        ...     @httpmethod
        ...     def BREW(self, req):
        ...         return Response("Putting the kettle on")
        ...     POST = BREW
        ...
        >>> print Request.blank("", method="BREW").get_response(HTCPTP())
        200 OK
        Content-Type: text/html; charset=UTF-8
        Content-Length: 21
        <BLANKLINE>
        Putting the kettle on
        >>> print Request.blank("", method="POST").get_response(HTCPTP())
        200 OK
        Content-Type: text/html; charset=UTF-8
        Content-Length: 21
        <BLANKLINE>
        Putting the kettle on
        >>> print Request.blank("", method="WHEN").get_response(HTCPTP())
        405 Method Not Allowed
        Content-Type: text/html; charset=UTF-8
        Content-Length: 0
        Allow: BREW, POST

    """
    @property
    def methods(self):
        """
        returns the list of httpmethods (by name) implemented for this
        resource.
        """
        _methods = []
        cls = type(self)
        for attr in dir(cls):
            if getattr(getattr(cls, attr), '_httpmethod', False):
                _methods.append(attr)
        return _methods

    def method_not_allowed(self, req):
        """
        Called when an HTTP method is specified in a request for which there is
        no defined httpmethod.  By default, this responds with a 405 Not
        Allowed response.
        """
        resp = Response(status=405)
        resp.allow = self.methods
        return resp

    @wsgify
    def __call__(self, req):
        if getattr(getattr(self, req.method, None), '_httpmethod', False):
            return getattr(self, req.method)
        try:
            super_call = super(HttpResource, self).__call__
        except AttributeError:
            return self.method_not_allowed(req)
        return super_call(req)



#! /usr/bin/env python

try:
    execfile
except NameError:
    import inspect

    def execfile(file, globals=None, locals=None):
        frame = None
        try:
            if globals is None and locals is None:
                frame = inspect.stack()[-2][0]
                globals = frame.f_globals
                locals = frame.f_locals
            elif globals is not None and locals is None:
                locals = globals
            elif globals is None and locals is not None:
                frame = inspect.stack()[-2][0]
                globals = frame.f_globals
        finally:
            del frame
        with open(file, "r") as fh:
            exec(fh.read() + "\n", globals, locals)

import ast
import time
import logging
import wsgiref.simple_server
import argparse
import threading

import autoreloader

PARSER = argparse.ArgumentParser()
PARSER.add_argument('-p', '--port', type=int, default=8000)
PARSER.add_argument('-a', '--address', type=str, default='')
PARSER.add_argument('-t', '--threads', type=int, default=10)
PARSER.add_argument('-r', '--autoreload', default=False, action='store_true')

PARSER.add_argument('--quiet', '-q',
                    default=logging.ERROR,
                    help='set log verbosity',
                    dest='loglevel',
                    action='store_const',
                    const=logging.CRITICAL)
PARSER.add_argument('--verbose', '-v',
                    dest='loglevel',
                    action='store_const',
                    const=logging.INFO)
PARSER.add_argument('--debug', '-d',
                    dest='loglevel',
                    action='store_const',
                    const=logging.DEBUG)
PARSER.add_argument('--logfile', dest='logfile', default=None)

PARSER.add_argument('app')


class AppMaker(object):
    """
    defers app loading until first request.
    """

    def __init__(self, app_string, autoreload=False):
        path, expr = app_string.split(':', 1)
        ast.parse(expr, 'app', mode='eval')
        self.app_string = app_string
        self._autoreload = autoreload
        self._lock = threading.Lock()

    @property
    def app(self):
        try:
            return self._application
        except AttributeError:
            with self._lock:
                # check *again* with the lock
                if hasattr(self, '_application'):
                    app = self._application
                else:
                    path, expr = self.app_string.split(':', 1)
                    try:
                        evalcontext = __import__(path, fromlist=['']).__dict__
                    except ImportError:
                        evalcontext = {'__name__': path}
                        execfile(path, evalcontext)
                        if self._autoreload:
                            autoreloader.assassinate([path])
                    self._application = app = eval(expr, evalcontext)
                return app

    def __call__(self, environ, start_response):
        app = self.app
        return app(environ, start_response)


def main():
    args = PARSER.parse_args()

    logging.basicConfig(filename=args.logfile, level=args.loglevel)
    if args.autoreload:
        autoreloader.assassinate([__file__])
    address = args.address
    port = args.port
    nthreads = max(1, args.threads)

    application = AppMaker(args.app, args.autoreload)
    if not args.autoreload:
        application = application.app

    server = wsgiref.simple_server.make_server(address, port, application)
    threads = [threading.Thread(target=server.serve_forever)
               for dummy in range(nthreads)]
    for thread in threads:
        thread.daemon = True
        thread.start()
    # don't join threads!
    print("Serving %r http://%s" % (server.application,
                                    "%s:%d" % server.server_address))

    try:
        while True:
            time.sleep(10000)
    except KeyboardInterrupt:
        print()
        print("Caught KeyboardInterrupt, bye!")
        return

if __name__ == '__main__':
    main()



"""
>>> from webob import Request, Response

"""
import copy
import mimetypes

from webob import exc
from webob.dec import wsgify


class vary(object):
    r"""

    decorator to support content negotiation.

        >>> from webob import Request, Response
        >>> import json, collections, csv, StringIO
        >>> class Foo(object):
        ...     FBBRow = collections.namedtuple('FBBRow', 'foo bar baz')
        ...     data = [FBBRow(1, 2, 3), FBBRow(4, 5, 6), FBBRow(7, 8, 9),]
        ...
        ...     @wsgify
        ...     def __call__(self, req):
        ...         return req.get_response(self.render(req))
        ...
        ...     @vary(default_media='text/html')
        ...     @wsgify
        ...     def render(self, req):
        ...         data = self.data
        ...         html = ("<html><head><title>%s</title></head>\n"
        ...                 "<body>\n%s\n</body></html>")
        ...         headings = "".join("<th>%s</th>" % label
        ...                            for label in self.FBBRow._fields)
        ...         tabledata = ["".join("<td>%d</td>" % field
        ...                              for field in row)
        ...                      for row in data]
        ...         table_rows = "\n".join("<tr>%s</tr>" % row
        ...                                for row in ([headings] + tabledata))
        ...         table = "<table>\n%s\n</table>" % table_rows
        ...         formatted = html % ("data!", table)
        ...         return formatted
        ...
        ...     @render(accept='text/csv')
        ...     @wsgify
        ...     def render(self, req):
        ...         data = self.data
        ...         outfile = StringIO.StringIO()
        ...         writer = csv.writer(outfile)
        ...         writer.writerow(self.FBBRow._fields)
        ...         writer.writerows(data)
        ...         return outfile.getvalue()
        ...
        ...     @render(accept='application/json')
        ...     @wsgify
        ...     def render(self, req):
        ...         data = self.data
        ...         return json.dumps(map(self.FBBRow._asdict, data))
        ...
        ...     @render(accept='application/javascript')
        ...     @wsgify
        ...     def render(self, req):
        ...         json_app = self.render.media['application/json']
        ...         jsondata = req.get_response(json_app)
        ...         try:
        ...             cb = req.GET['jsonp_callback']
        ...             resp = Response()
        ...             resp.text = u"%s(%s)" % (cb, jsondata.text)
        ...             return resp
        ...         except KeyError:
        ...             return jsondata
        ...

    When the client doesn't specify a content type, the default is used.

        >>> req = Request.blank('')
        >>> print req.get_response(Foo())
        200 OK
        Content-Type: text/html; charset=UTF-8
        Content-Length: 244
        Vary: accept
        <BLANKLINE>
        <html><head><title>data!</title></head>
        <body>
        <table>
        <tr><th>foo</th><th>bar</th><th>baz</th></tr>
        <tr><td>1</td><td>2</td><td>3</td></tr>
        <tr><td>4</td><td>5</td><td>6</td></tr>
        <tr><td>7</td><td>8</td><td>9</td></tr>
        </table>
        </body></html>

    Accept headers are honored:

        >>> req.accept = 'application/json'
        >>> print req.get_response(Foo())   # doctest: +NORMALIZE_WHITESPACE
        200 OK
        Content-Length: 96
        Content-Type: application/json; charset=UTF-8
        Vary: accept
        <BLANKLINE>
        [{"foo": 1, "bar": 2, "baz": 3},
         {"foo": 4, "bar": 5, "baz": 6},
         {"foo": 7, "bar": 8, "baz": 9}]
        >>> req.accept = 'text/csv'
        >>> print req.get_response(Foo())   # doctest: +NORMALIZE_WHITESPACE
        200 OK
        Content-Length: 34
        Content-Type: text/csv; charset=UTF-8
        Vary: accept
        <BLANKLINE>
        foo,bar,baz
        1,2,3
        4,5,6
        7,8,9

    There's nothing wrong with using one handler to support another:

        >>> req = Request.blank('?jsonp_callback=gotData')
        >>> req.accept = 'application/javascript'
        >>> print req.get_response(Foo())   # doctest: +NORMALIZE_WHITESPACE
        200 OK
        Content-Length: 105
        Content-Type: application/javascript; charset=UTF-8
        Vary: accept
        <BLANKLINE>
        gotData([{"foo": 1, "bar": 2, "baz": 3},
                 {"foo": 4, "bar": 5, "baz": 6},
                 {"foo": 7, "bar": 8, "baz": 9}])

    Real HTTP clients seldom set the HTTP Accept header correctly; standard
    practice is to use a filename or other media type distinguishing means in
    the url.

        >>> Request.blank('').get_response(Foo()).content_type
        'text/html'
        >>> Request.blank('?filename=data.csv').get_response(Foo()).content_type
        'text/csv'
        >>> Request.blank('?accept=application/json').get_response(Foo()).content_type
        'application/json'

    """

    def __init__(self, default_media='text/html'):
        self.default_media = default_media
        self.functions = {}

    def __call__(self, wrapped=None, accept=None):
        clone = copy.deepcopy(self)

        if accept is None:
            accept = clone.default_media

        def decorate(method):
            clone.functions[accept] = method
            return clone

        if wrapped:
            return decorate(wrapped)
        else:
            return decorate

    def __get__(self, instance, owner):
        # FIXME: This seems inefficient... oh well.
        if instance is None:
            return self
        vary_app = VaryResponse()
        vary_app.default_media = self.default_media
        for media, function in self.functions.iteritems():
            if hasattr(function, '__get__'):
                function = function.__get__(instance, owner)
            vary_app.media[media] = function
        return vary_app


class VaryResponse(object):
    """
    content negotiation capable wsgi app.  able to return client's choice of
    media types.  see vary for details.  End users should not need to make
    instances of or subclass this class
    """
    default_media = None

    @property
    def media(self):
        """
        a dict mapping uri's to wsgi applications.
        """
        try:
            collection = self.__collection
        except AttributeError:
            self.__collection = collection = {}
        return collection

    @wsgify
    def no_accept(self, req):
        """
        called when the request did not specify an 'Accept' header.  The
        request will be retried with accept of self.default_media
        """
        if req.GET.get('filename'):
            filename = req.GET['filename']
            accept = mimetypes.guess_type(filename)[0]
            if accept:
                req.accept = accept
                return
        try:
            default_media = self.default_media
        except AttributeError:
            return
        req.accept = default_media

    @wsgify
    def no_match(self, req):
        """
        """
        raise exc.HTTPPreconditionFailed

    def media_mismatch(self, req, match, resp):
        """
        called when the returned response did not satisfy the accepted media
        type.   If the resp has a media type of 'text/html', it will just be
        changed to the matched media type (the handler probably forgot to set
        it), and raises a HTTPPreconditionFailed otherwise.
        """
        # XXX
        if resp.content_type == 'text/html':
            resp.content_type = match
            return resp
        else:
            raise exc.HTTPPreconditionFailed

    @wsgify
    def __call__(self, req):
        #
        if req.GET.get('accept'):
            req.accept = req.GET['accept']
        if not req.accept or req.accept == '*/*':
            mimetype, _ = mimetypes.guess_type(req.path)
            if mimetype and mimetype in self.media:
                req.accept = mimetype
            else:
                self.no_accept(req)

        medias = self.media.keys()
        #medias.sort(key=(self.default_media.__ne__))
        match = req.accept.best_match(medias)
        # TODO: if no accept header (or generic accept:text/html), also
        # try using guess_type on the requested url.

        if match and req.accept.quality(match):
            resp = req.get_response(self.media[match])
            # "cheat a little...", match, resp.content_type
            if match != resp.content_type == 'text/html':
                resp.content_type = match
            elif not(req.accept.quality(resp.content_type)):
                resp = self.media_mismatch(req, match, resp)

        else:
            resp = req.get_response(self.no_match)

        # cache control:
        if self.media:
            resp.vary = (resp.vary or ()) + ('accept',)
        # TODO: localize
        #if self.locale:
        #    resp.vary = (resp.vary or ()) + ('accept-language',)

        return resp



from webob import exc
import wsgiref.util


def all_sub_paths(environ):
    """
    return a list containing the pairs of script_name and path_info that the
    supplied environ can contain, from shortest script name to longest.  Note
    this will also include the 'root', with empty script_name.  the existing
    environ is not modified.

    >>> all_sub_paths({'PATH_INFO': '/foo/bar'})
    [('', '/foo/bar'), ('/foo', '/bar'), ('/foo/bar', '')]
    >>> all_sub_paths({'PATH_INFO': '/foo/bar/'})
    [('', '/foo/bar/'), ('/foo', '/bar/'), ('/foo/bar', '/'), ('/foo/bar/', '')]
    >>> all_sub_paths({'PATH_INFO': '/foo//bar'})
    [('', '/foo//bar'), ('/foo', '/bar'), ('/foo/bar', '')]

    """
    test_environ = {'SCRIPT_NAME': '',
                    'PATH_INFO': environ['PATH_INFO']}
    script_names = []
    script_names.append((test_environ['SCRIPT_NAME'],
                         test_environ['PATH_INFO']))
    while wsgiref.util.shift_path_info(test_environ) is not None:
        script_names.append((test_environ['SCRIPT_NAME'],
                             test_environ['PATH_INFO']))
    return script_names


class ScriptCollection(object):
    r"""
    organize a pile of applications by "script_name".

    Mount some applications at assorted uri's

        >>> from webob import Request, Response
        >>> collection = ScriptCollection.from_dict(
        ...     {'': Response(body="Welcome"),
        ...      '/beer': exc.HTTPUnauthorized("21 and Over"),
        ...      '/foo/bar/baz/long/path': Response("That was tough!")})

    matching names are dispatched to the named app

        >>> print Request.blank("").get_response(collection)
        200 OK
        Content-Type: text/html; charset=UTF-8
        Content-Length: 7
        <BLANKLINE>
        Welcome
        >>> print Request.blank("/").get_response(collection)
        200 OK
        Content-Type: text/html; charset=UTF-8
        Content-Length: 7
        <BLANKLINE>
        Welcome
        >>> print Request.blank("/beer").get_response(collection)
        ...     # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        401 Unauthorized
        Content-Length: 264
        Content-Type: text/plain; charset=UTF-8
        <BLANKLINE>
        401 Unauthorized
        <BLANKLINE>
        This server could not verify ...
        <BLANKLINE>
        21 and Over
        >>> print Request.blank("/foo/bar").get_response(collection)
        200 OK
        Content-Type: text/html; charset=UTF-8
        Content-Length: 7
        <BLANKLINE>
        Welcome
        >>> print Request.blank("/foo/bar/baz/long/path/quux"
        ...                     ).get_response(collection)
        200 OK
        Content-Type: text/html; charset=UTF-8
        Content-Length: 15
        <BLANKLINE>
        That was tough!

    """

    @classmethod
    def from_dict(cls, apps):
        self = cls()
        for script_name, app in apps.iteritems():
            self.scripts[script_name] = app
        return self

    @property
    def scripts(self):
        """
        a dict mapping uri's to wsgi applications.
        """
        try:
            collection = self.__collection
        except AttributeError:
            self.__collection = collection = {}
        return collection

    def __call__(self, environ, start_response):
        for (script_name, path_info) in reversed(all_sub_paths(environ)):
            if script_name in self.scripts:
                environ['SCRIPT_NAME'] += script_name
                environ['PATH_INFO'] = path_info
                return self.scripts[script_name](environ, start_response)
        try:
            super_call = super(ScriptCollection, self).__call__
        except AttributeError:
            return exc.HTTPNotFound()(environ, start_response)
        return super_call(environ, start_response)



import collections
import datetime
import mimetypes
import os

import jinja2
import texttable


from webob.dec import wsgify

from wsgilite.apps import static
from wsgilite.extras.restructured import RstResource
from wsgilite.framework.method import HttpResource

fancy_index_item = collections.namedtuple('fancy_index_item', (
    'filename',
    'link_path',
    'isdir',
    'content_type',
    'content_encoding',
    'content_length',
    'date'))


@jinja2.contextfunction
def tabulate(ctx, template_name, items, header=None):
    template = ctx.environment.get_template(template_name)
    table = texttable.Texttable(0)
    if not header:
        header = template.blocks.keys()
    table.header(header)
    for item in items:
        item_ctx = template.new_context(ctx)
        item_ctx.vars['row'] = item
        row = []
        for field in header:
            row.append(''.join(template.blocks[field](item_ctx)))
        table.add_row(row)
    return table.draw()


ENV = jinja2.Environment(
    loader=jinja2.PackageLoader(__name__, 'templates'))

ENV.globals['tabulate'] = tabulate


class FancyIndex(RstResource, HttpResource):
    env = ENV
    show_hidden = False

    def __init__(self, fsdir, baseurl='/'):
        self.fsdir = fsdir
        self.baseurl = baseurl

    @wsgify
    def GET(self, req):
        return req.get_response(self.render)

    @RstResource.render(accept='text/x-rst')
    @wsgify
    def render(self, req):
        return self.env.get_template('index.rst').render(
            fsdir=self.fsdir,
            baseurl=req.script_name.rstrip('/') + self.baseurl,
            items=self.getdirlist())

    def getdirlist(self):
        #dirlist_path = os.path.join(self.path, self.baseurl)
        dirlist = [fancy_index_item(filename='.',
                                    link_path=self.baseurl,
                                    isdir=True,
                                    content_type=None,
                                    content_encoding=None,
                                    content_length=0,
                                    date=None,
                                    ),
                   fancy_index_item(filename='..',
                                    link_path=os.path.dirname(self.baseurl),
                                    isdir=True,
                                    content_type=None,
                                    content_encoding=None,
                                    content_length=0,
                                    date=None,
                                    ), ]
        for dirent in os.listdir(self.fsdir):
            if not self.show_hidden and dirent.startswith('.'):
                continue
            dirent_path = os.path.join(self.fsdir, dirent)
            content_type, content_encoding = mimetypes.guess_type(dirent)
            stat_info = os.stat(dirent_path)
            dirlist.append(fancy_index_item(filename=dirent,
                                            link_path='/'.join((self.baseurl.rstrip('/'), dirent)),
                                            isdir=os.path.isdir(dirent_path),
                                            content_type=content_type,
                                            content_encoding=content_encoding,
                                            content_length=stat_info.st_size,
                                            date=datetime.datetime.fromtimestamp(stat_info.st_mtime)))

        dirlist.sort()
        return dirlist


class FancyDir(static.StaticDir):
    index_app = FancyIndex



import webob
import os
import unittest
from wsgilite.apps import static

FIXTURES = os.path.join(os.path.dirname(__file__), 'fixtures')


class TestStatic(unittest.TestCase):
    def test_static_file(self):
        resp = webob.Request.blank('').get_response(static.StaticFile(os.path.join(FIXTURES, 'index.html')))
        self.assertEquals(resp.status_int, 200)
        self.assertEquals(resp.content_type, 'text/html')



import os
#from distutils.core import setup
from setuptools import setup

PACKAGES = [
    'wsgilite',
    'wsgilite/apps',
    'wsgilite/config',
    'wsgilite/extras',
    'wsgilite/framework',
]

PACKAGE_DATA = {'wsgilite/extras': ['templates/*.rst']}


def walkfunc(arg, dir, files):
    if '__init__.py' in files:
        arg.add(dir)


if os.path.isdir('.git'):
    foundpackages = set()
    os.path.walk('wsgilite', walkfunc, foundpackages)
    if foundpackages != set(PACKAGES):
        print "Expected"
        print sorted(list(set(PACKAGES)))
        print "Found"
        print sorted(list(foundpackages))
        print "This looks funny!"


setup(
    name='wsgilite',
    version='0.1.0',
    author='dan bornside',
    author_email='dan.bornside@gmail.com',
    description='a grab-bag of web app utilities for developer convenience',
    long_description=open('README.rst').read(),
    url='https://github.com/danbornside/wsgilite',
    packages=PACKAGES,
    package_data=PACKAGE_DATA,
    entry_points={
        'console_scripts': [
            'wsgilite = wsgilite.wsgilite:main'
        ]},
    install_requires=[
        'Jinja2',
        'Pygments',
        'WebOb',
        'docutils',
        'texttable',
    ]
)
